#!/bin/sh
# the next line restarts using wish \
exec plotwish -f "$0" "$@"

proc error_exit {msg} {
  wm withdraw .
  tk_dialog .error Error $msg error 0 Exit
  exit 1
}

if {[catch {package require Tk 8.0} msg]} {
  error_exit $msg
}

#----- get startup directory and name

set cmd_name [file tail $argv0]
set cmd_dir  [file dirname $argv0]
if {![file exists $argv0] || [file isdirectory $argv0]} {
  if {$tcl_platform(platform) == "windows"} {
    set sep ";"
  } else {
    set sep ":"
  }
  foreach i [split $env(PATH) $sep] {
    if {$sep == ";"} {
      set i [join [split $i \\] /]
    }
    if {[file exists $i/$cmd_name] && ![file isdirectory $i/$cmd_name]} {
      set cmd_dir $i
      break;
    }
  }
}
set curdir [pwd]
if ![catch {cd $cmd_dir}] {
  set cmd_dir [pwd]
  cd $curdir
}
if {$tcl_platform(platform) == "windows"} {
  set cmd_dir [file attributes $cmd_dir -shortname]
}

#----- set path to tcl scripts

set auto_path "$cmd_dir $cmd_dir/../common $auto_path"
if {[info exists env(TCL_PROC_DIR)]} {
  set auto_path "$env(TCL_PROC_DIR) $auto_path"
}

#----- initialize

if {$tcl_platform(platform) == "windows"} {
  set vers [join [split $tcl_version .] {}]
  if {[info commands CGNSopen] == {}} {
    if {[catch {load cgnstcl$vers} msg]} {
      error_exit $msg
    }
  }
  catch {load tclreg$vers registry}
} else {
  if {[info commands CGNSopen] == {}} {
    error_exit "need to run script with plotwish"
  }
}

if {[catch {package require Tkogl} msg]} {
  error_exit $msg
}

if [catch {config_defaults 1} msg] {error_exit $msg}

array set ProgData {
  cgnsfile ""
  displaylist ""
  axislist ""
  background {0.0 0.0 0.0}
  colors { \
    {0.8 0.8 0.8} \
    {1.0 0.0 0.0} \
    {1.0 1.0 0.0} \
    {0.0 1.0 0.0} \
    {0.0 1.0 1.0} \
    {0.0 0.0 1.0} \
    {1.0 0.0 1.0} \
  }
  twosided 0
  culling disable
  axis 0
  meshvis 1
  meshmode 1
  regvis 0
  regmode 2
  edgeangle 60
  fov 30
  np 0.025
  fp 2.0
  nzones 0
  zones {}
  curzone ""
  curreg ""
  zoneclr ""
  zonevis 0
  regiondim 0
  regionclr ""
  regionvis 0
  regionmode 0
  badregion \#FF5050
  reg,file ".cgnstools"
  reg,base "HKEY_CURRENT_USER/Software/CGNS"
  reg,key  "CGNSplot"
  reg,vals {cgnsfile background colors twosided culling axis \
            edgeangle meshvis meshmode regvis regmode fov}
}

#----- scan for startup options

set oglargs {}
set oglwidth 500
set oglheight 500
foreach {opt value} $argv {
  switch -glob -- $opt {
    -dep* {lappend oglargs -depth $value}
    -dou* {lappend oglargs -doublebuffer $value}
    -wid* {set oglwidth $value}
    -hei* {set oglheight $value}
  }
}

#----- read registry

if {[tclreg_init -base $ProgData(reg,base) -fname $ProgData(reg,file)]} {
  foreach i $ProgData(reg,vals) {
    if {![catch {tclreg_get $ProgData(reg,key) $i} val] && $val != ""} {
      set ProgData($i) $val
    }
  }
}

#---------- main window

wm title . "CGNSplot"
wm protocol . WM_DELETE_WINDOW do_quit

proc do_quit {} {
  global ProgData
  catch CGNSclose
  foreach i $ProgData(reg,vals) {
    catch {tclreg_set $ProgData(reg,key) $i $ProgData($i)}
  }
  catch tclreg_close
  catch {WinHtml close}
  destroy .
  exit 0
}

#----- menu

menubar_create {File Display Help}

#----- file menu

set m [menubar_get File]
$m add command -label "Load CGNS..." -command load_cgns
$m add separator
$m add command -label "Quit" -command do_quit

#----- display menu

set m [menubar_get Display]
$m add command -label "Reset View" -command {reset_view; $OGLwin redraw}
$m add command -label "Center View" -command Center
$m add command -label "Set Defaults..." -command set_defaults
$m add command -label "Set Perspective..." -command set_perspective
$m add separator
$m add checkbutton -label "Two Sided" -command set_twosided \
  -variable ProgData(twosided) -onvalue 1 -offvalue 0
$m add checkbutton -label "Backface Culling" -command set_culling \
  -variable ProgData(culling) -onvalue enable -offvalue disable
$m add checkbutton -label "Display Axis" -command set_axis \
  -variable ProgData(axis) -onvalue 1 -offvalue 0

proc set_twosided {} {
  global ProgData OGLwin
  if {$ProgData(twosided)} {
    .toolbar.but.b3.twosided configure -relief sunken
    set_balloon .toolbar.but.b3.twosided "Disable Two-Sided Lighting"
  } else {
    .toolbar.but.b3.twosided configure -relief raised
    set_balloon .toolbar.but.b3.twosided "Enable Two-Sided Lighting"
  }
  $OGLwin eval -lightmodel lightmodeltwoside $ProgData(twosided)
  $OGLwin redraw
}

proc set_culling {} {
  global ProgData OGLwin
  if {$ProgData(culling) == "enable"} {
    .toolbar.but.b3.culling configure -relief sunken
    set_balloon .toolbar.but.b3.culling "Disable Backface Culling"
  } else {
    .toolbar.but.b3.culling configure -relief raised
    set_balloon .toolbar.but.b3.culling "Enable Backface Culling"
  }
  $OGLwin eval -$ProgData(culling) cullface
  $OGLwin redraw
}

proc set_axis {} {
  global ProgData OGLwin
  if {$ProgData(axis)} {
    .toolbar.but.b3.axis configure -relief sunken
    set_balloon .toolbar.but.b3.axis "Disable Axis Display"
  } else {
    .toolbar.but.b3.axis configure -relief raised
    set_balloon .toolbar.but.b3.axis "Enable Axis Display"
  }
  OGLaxis $ProgData(axis)
  $OGLwin redraw
}

#----- help menu

set m [menubar_get Help]
$m add command -label "CGNSplot Help..." -command {help_show cgnsplot}
$m add command -label "CGNS Help..." -command {help_show cgns}
$m add separator
$m add command -label "Configure Help..." -command help_setup
$m add separator
$m add command -label "About..." -underline 0 -command do_about

image create photo img_about -data {\
R0lGODlhIAAgALMAAAAA/wD//4QAAMbGxv8AAP//////////////////////////////////\
/////////ywAAAAAIAAgAAAEv3AMIGelEttbN8+fp3mkGGpj2plgawluiZ6XAJMooO98v9sC\
H881cwkINtyqyBocgaoOs3hE2qI0oqQKvMWas6o1eRKaxd2gT0sUj8my7MSdvo3AFfebfJ/m\
CYCAdTA5Zj+BgYM9AQEAjI6MjYx6e1CRko+RkJSVNpiYkJmTiIJ1mp+hnwE2iHWphmeJXUKi\
p5eSXEC2l6mqorO3r7e+tQK1tb3ExMmOsM7Ow8rSydHU09HD1tfYx9nYvc/hQhEAADs=}

proc do_about {} {
  global ProgData
  dialog .about -1 -1 "About CGNSplot" \
"CGNSplot Version 2.4

ICEM CFD Engineering
2855 Telegraph Ave., Suite 501
Berkeley, CA. 94705

Bruce Wedan
bruce.wedan@ansys.com" img_about 0 Close
}

#---------- toolbar

frame .toolbar
pack .toolbar -side top -pady 2 -fill x

set f [frame .toolbar.but]
pack $f -side left

#---

image create photo img_open -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwQ4EMlJKwJvZcC7BxdnSV04nCgKjtR6vZgmZ49L\
bmlus7xV9j4QQPYRtWbI3RCXU10WgKaTVPQAexEAOw==}

button $f.open -image img_open -takefocus 0 \
  -command {load_cgns ""}
pack $f.open -side left -padx 5
set_balloon $f.open "Open a CGNS File..."

#---

image create photo img_reset -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAAAQiEMlJKwXYWqA17xXGjQgZguA3qWhbni58vaQY\
vzHb2XdvRQA7}

image create photo img_center -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwQmEMmJAL22Ypmp/dolduJYcqiXqiB5umcFwG9s\
YzOd07fZy7yfJAIAOw==}

image create photo img_colors -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwRFEMlJ63TYsc2ab1QiJku5KKhCPewDvMAhH+Fo\
nimVadwHTi0XbEazGI/IyohkSqkmy5tTl+F0PlSM1bdqwWKzLutLTB4jADs=}

image create photo img_viewing -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwQqEMlJqwT2uo0z6JkFht44fdSnVuipmiXpyUhL\
wiwMbA5onzzcj3bB0ToRADs=}

set b [frame $f.b2]
pack $b -side left -padx 5

button $b.reset -image img_reset -takefocus 0 \
  -command {reset_view; $OGLwin redraw}
set_balloon $b.reset "Reset View"

button $b.center -image img_center -takefocus 0 \
  -command Center
set_balloon $b.center "Center View"

button $b.colors -image img_colors -takefocus 0 \
  -command set_defaults
set_balloon $b.colors "Set Defaults..."

button $b.viewing -image img_viewing -takefocus 0 \
  -command set_perspective
set_balloon $b.viewing "Set Perspective..."

pack $b.reset $b.center $b.colors $b.viewing -side left -padx 1

#---

image create photo img_twosided -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwRGEMlJKwLnUMZoBtSyPNwjWpJIMuaCbh2KTQvn\
Sgd4aak95S/OyyIc0nyyG6KIWOgwB9GoJP1ApavW4iMBKJnOV42hNDIpEQA7}

image create photo img_culling -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAAAQqEMlJq5UgX6oxuF0VhuCHfKR3YlM3spzZwpxI\
z/Vq56nc+zcRKlXaGC0RADs=}

image create photo img_axis -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwQwEEkJUJ04o5V6p9o2hJlAnteJLaOKKIqJujNt\
Z+ydY1UqwYKAMGAp/mIqgNKm9J0iADs=}

set b [frame $f.b3]
pack $b -side left -padx 5

button $b.twosided -image img_twosided -takefocus 0 \
  -command toggle_twosided
if {$ProgData(twosided)} {
  $b.twosided configure -relief sunken
  set_balloon $b.twosided "Disable Two-Sided Lighting"
} else {
  set_balloon $b.twosided "Enable Two-Sided Lighting"
}

button $b.culling -image img_culling -takefocus 0 \
  -command toggle_culling
if {$ProgData(culling) == "enable"} {
  $b.culling configure -relief sunken
  set_balloon $b.culling "Disable Backface Culling"
} else {
  set_balloon $b.culling "Enable Backface Culling"
}

button $b.axis -image img_axis -takefocus 0 \
  -command toggle_axis
if {$ProgData(axis)} {
  $b.axis configure -relief sunken
  set_balloon $b.axis "Disable Axis Display"
} else {
  set_balloon $b.axis "Enable Axis Display"
}

pack $b.twosided $b.culling $b.axis -side left -padx 1

proc toggle_twosided {} {
  global ProgData
  if {$ProgData(twosided)} {
    set ProgData(twosided) 0
  } else {
    set ProgData(twosided) 1
  }
  set_twosided
}

proc toggle_culling {} {
  global ProgData
  if {$ProgData(culling) == "enable"} {
    set ProgData(culling) disable
  } else {
    set ProgData(culling) enable
  }
  set_culling
}

proc toggle_axis {} {
  global ProgData
  if {$ProgData(axis)} {
    set ProgData(axis) 0
  } else {
    set ProgData(axis) 1
  }
  set_axis
}

#---

image create photo img_help -data {\
R0lGODlhEAAQALMAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A\
/wD//////yH5BAEAAAgALAAAAAAQABAAQwQiEMlJq50kX5kJ1hvShd+4mSJ4qmTrXl28ehw7\
t+j75joVAQA7}

button $f.help -image img_help -takefocus 0 \
  -command {help_show cgnsplot}
set_balloon $f.help "CGNSplot Help..."

pack $f.help -side left -padx 5

#----- status bar

set ProgData(status) ""
label .status -textvariable ProgData(status) -relief sunken -anchor w
pack .status -side bottom -fill x -padx 3 -pady 3

proc display_message {msg} {
  global ProgData
  set ProgData(status) $msg
  update idletasks
}

#----- OpenGL window

frame .plot -relief sunken -bd 2
pack .plot -side left -fill both -expand 1 -padx 3

set OGLwin .plot.gl

if {[catch {eval OGLwin $OGLwin -width $oglwidth \
  -height $oglheight $oglargs} msg]} {
  error_exit $msg
}
pack $OGLwin -side left -fill both -expand 1

#----- selection window

frame .sel
pack .sel -side right -fill y

#--- base selection

set f [frame .sel.base]
pack $f -side top -padx 5 -fill x
label $f.lab -text Base
pack $f.lab -side top
set ProgData(baselist) [ComboboxCreate $f.cb -width 10 \
  -edit 0 -command set_base]
pack $f.cb -side top -fill x
bind $f.cb.ent <ButtonRelease-3> {base_properties %W %X %y}

#--- zone selection

set f [frame .sel.zone]
pack $f -side top -padx 5 -fill x
label $f.lab -text Zones
pack $f.lab -side top
set ff [frame $f.list]
pack $ff -side top -fill x
scrollbar $ff.ys -command "$ff.list yview"
pack $ff.ys -side right -fill y
scrollbar $ff.xs -orient horizontal -command "$ff.list xview"
pack $ff.xs -side bottom -fill x
set ProgData(zonelist) [listbox $ff.list -width 20 -height 5 \
  -yscroll "$ff.ys set" -xscroll "$ff.xs set" -exportselection 0]
pack $ff.list -side left -fill x -expand 1

bind $ff.list <ButtonRelease-1> {set_zone [%W nearest %y]}
bind $ff.list <Double-Button-1> {toggle_zone [%W nearest %y]}
bind $ff.list <ButtonRelease-3> {zone_properties %W %X %y}

set f [frame .sel.zopt]
pack $f -side top -fill x -padx 5 -pady 3
checkbutton $f.w -text Visible -variable ProgData(zonevis) \
  -onvalue 1 -offvalue 0 -command update_zone -state disabled
pack $f.w -side left
set ProgData(zonebut) [button $f.c -relief solid -text Color... \
  -borderwidth 1 -command "zone_color $f.c" -state disabled]
pack $f.c -side right

frame .sel.sep1 -bd 1 -height 3 -relief sunken
pack .sel.sep1 -side top -padx 5 -fill x

proc zone_state {state} {
  foreach i {w c} {
    .sel.zopt.$i configure -state $state
  }
}

#--- region selection

set f [frame .sel.reg]
pack $f -side top -padx 5 -fill both -expand 1
label $f.lab -text Regions
pack $f.lab -side top
set ff [frame $f.list]
pack $ff -side top -fill both -expand 1
scrollbar $ff.ys -command "$ff.list yview"
pack $ff.ys -side right -fill y
scrollbar $ff.xs -orient horizontal -command "$ff.list xview"
pack $ff.xs -side bottom -fill x
set ProgData(reglist) [listbox $ff.list -width 20 -height 5 \
  -yscroll "$ff.ys set" -xscroll "$ff.xs set" \
  -exportselection 0 -selectmode extended]
pack $ff.list -side left -fill both -expand 1

bind $ff.list <ButtonRelease-1> {set_region [%W curselection]}
bind $ff.list <Double-Button-1> {toggle_region [%W nearest %y]}
bind $ff.list <ButtonRelease-3> {region_properties %W %X %y}

set f [frame .sel.ropts]
pack $f -side top -fill x -padx 5 -pady 3
checkbutton $f.w -text Visible -variable ProgData(regionvis) \
  -onvalue 1 -offvalue 0 -command update_region -state disabled
pack $f.w -side left
set ProgData(regionbut) [button $f.c -relief solid -text Color... \
  -borderwidth 1 -command "region_color $f.c" -state disabled]
pack $f.c -side right

frame .sel.sep2 -bd 1 -height 3 -relief sunken
pack .sel.sep2 -side top -padx 5 -fill x

frame .sel.f2
pack .sel.f2 -side top -fill x -padx 5

set f [frame .sel.f2.l]
pack $f -side left
radiobutton $f.o -text "Outline" -variable ProgData(regionmode) -value 1 \
   -command update_region -state disabled
radiobutton $f.f -text "Flat" -variable ProgData(regionmode) -value 3 \
   -command update_region -state disabled
pack $f.o $f.f -side top -anchor w

set f [frame .sel.f2.r]
pack $f -side right
radiobutton $f.m -text "Mesh" -variable ProgData(regionmode) -value 2 \
   -command update_region -state disabled
radiobutton $f.s -text "Smooth" -variable ProgData(regionmode) -value 4 \
   -command update_region -state disabled
pack $f.m $f.s -side top -anchor w

proc region_state {dim} {
  if {$dim} {
    set state normal
  } else {
    set state disabled
  }
  foreach i {w c} {
    .sel.ropts.$i configure -state $state
  }
  foreach i {l.o l.f r.m r.s} {
    .sel.f2.$i configure -state $state
  }
}

#----- help

proc help_menu {} {
  if {[help_valid cgnsplot]} {
    menubar_state Help normal 0
    .toolbar.but.help configure -state normal
  } else {
    menubar_state Help disabled 0
    .toolbar.but.help configure -state disabled
  }
  if {[help_valid cgns]} {
    menubar_state Help normal 1
  } else {
    menubar_state Help disabled 1
  }
}

help_init {cgnsplot CGNSplot} {cgns CGNS}

#----- set default colors and zone/region viewing

proc set_defaults {{loc .}} {
  global ProgData Defaults OGLwin fgColor Font

  foreach i {meshvis meshmode regvis regmode} {
    set Defaults($i) $ProgData($i)
  }
  set Defaults(0) $ProgData(background)
  set nc 0
  foreach c $ProgData(colors) {
    incr nc
    set Defaults($nc) $c
  }
  set Defaults(nc) $nc

  set w .defaults
  toplevel $w
  wm title $w "Defaults"
  wm transient $w .
  wm protocol $w WM_DELETE_WINDOW {set ProgData(done) 0}

  FrameCreate $w.zone -text "Volume Mesh" -font $Font(bold) -padx 0 -pady 0
  pack $w.zone -side top -padx 5 -pady 2
  set zone [FrameGet $w.zone]

  checkbutton $zone.vis -text "Visible as" -onvalue 1 -offvalue 0 \
    -variable Defaults(meshvis)
  pack $zone.vis -side left
  set f [frame $zone.m1]
  radiobutton $f.o -text "Outline" -variable Defaults(meshmode) -value 1
  radiobutton $f.f -text "Flat" -variable Defaults(meshmode) -value 3
  pack $f.o $f.f -side top -anchor w
  set f [frame $zone.m2]
  radiobutton $f.m -text "Mesh" -variable Defaults(meshmode) -value 2
  radiobutton $f.s -text "Smooth" -variable Defaults(meshmode) -value 4
  pack $f.m $f.s -side top -anchor w
  pack $zone.m2 $zone.m1 -side right

  FrameCreate $w.reg -text Regions -font $Font(bold) -padx 0 -pady 0
  pack $w.reg -side top -padx 5 -pady 2
  set reg [FrameGet $w.reg]

  checkbutton $reg.vis -text "Visible as" -onvalue 1 -offvalue 0 \
    -variable Defaults(regvis)
  pack $reg.vis -side left
  set f [frame $reg.m1]
  radiobutton $f.o -text "Outline" -variable Defaults(regmode) -value 1
  radiobutton $f.f -text "Flat" -variable Defaults(regmode) -value 3
  pack $f.o $f.f -side top -anchor w
  set f [frame $reg.m2]
  radiobutton $f.m -text "Mesh" -variable Defaults(regmode) -value 2
  radiobutton $f.s -text "Smooth" -variable Defaults(regmode) -value 4
  pack $f.m $f.s -side top -anchor w
  pack $reg.m2 $reg.m1 -side right

  FrameCreate $w.colors -text Colors -font $Font(bold)
  pack $w.colors -side left -padx 5 -pady 2 -anchor n
  set colors [FrameGet $w.colors]

  set f [frame $colors.c0]
  pack $f -side top
  label $f.lab -text Background -width 10 -anchor w
  button $f.but -text Select -relief solid -bd 1 \
    -command "set_color $f.but 0"
  color_button $f.but $Defaults(0)
  pack $f.lab $f.but -side left -padx 5

  for {set n 1} {$n <= $nc} {incr n} {
    set f [frame $colors.c$n]
    pack $f -side top -pady 2
    label $f.lab -text "Color $n" -width 10 -anchor w
    button $f.but -text Select -relief solid -bd 1 \
      -command "set_color $f.but $n"
    color_button $f.but $Defaults($n)
    pack $f.lab $f.but -side left -padx 5
  }

  set f [frame $w.but]
  pack $f -side right -padx 5 -pady 2 -anchor n
  button $f.default -text Defaults -width 8 -command "default_values $w"
  button $f.add -text Add -width 8 -command "add_color $w"
  button $f.remove -text Remove -width 8 -command "remove_color $w"
  button $f.accept -text Accept -width 8 -command {set ProgData(done) 1}
  button $f.cancel -text Cancel -width 8 -command {set ProgData(done) 0}
  pack $f.default $f.add $f.remove $f.accept $f.cancel -side top -pady 5

  center_window $w $loc
  set oldFocus [focus]
  set oldGrab [grab current $w]
  if {$oldGrab != ""} {
    set grabStatus [grab status $oldGrab]
  }
  catch {grab $w}
  tkwait visibility $w
  focus $w
  tkwait variable ProgData(done)
  catch {focus $oldFocus}
  destroy $w
  if {$oldGrab != ""} {
    if {$grabStatus == "global"} {
      grab -global $oldGrab
    } else {
      grab $oldGrab
    }
  }

  if {$ProgData(done)} {
    foreach i {meshvis meshmode regvis regmode} {
      set ProgData($i) $Defaults($i)
    }
    set ProgData(background) $Defaults(0)
    set ProgData(colors) ""
    for {set n 1} {$n <= $Defaults(nc)} {incr n} {
      lappend ProgData(colors) $Defaults($n)
    }
    eval $OGLwin eval -clearcolor $ProgData(background)
    init_display
    update_zone
    set_region $ProgData(curreg)
    Center
  }
}

proc default_values {w} {
  global Defaults
  if {$Defaults(nc) != 7} {
    set cnt [expr 7 - $Defaults(nc)]
    if {$cnt < 0} {
      remove_color $w [expr abs($cnt)]
    } else {
      add_color $w $cnt
    }
  }
  array set Defaults {
    meshvis 1
    meshmode 1
    regvis 0
    regmode 2
    nc 7
    0 {0.0 0.0 0.0}
    1 {0.8 0.8 0.8}
    2 {1.0 0.0 0.0}
    3 {1.0 1.0 0.0}
    4 {0.0 1.0 0.0}
    5 {0.0 1.0 1.0}
    6 {0.0 0.0 1.0}
    7 {1.0 0.0 1.0}
  }

  set cf [FrameGet $w.colors]
  for {set n 0} {$n <= 7} {incr n} {
    color_button $cf.c$n.but $Defaults($n)
  }
}

proc add_color {w {cnt 1}} {
  global Defaults
  set n $Defaults(nc)
  incr Defaults(nc) $cnt
  set cf [FrameGet $w.colors]
  while {$n < $Defaults(nc)} {
    incr n
    set Defaults($n) {0.5 0.5 0.5}
    set f [frame $cf.c$n]
    pack $f -side top -pady 2
    label $f.lab -text "Color $n" -width 10 -anchor w
    button $f.but -text Select -relief solid -bd 1 \
      -command "set_color $f.but $n"
    color_button $f.but $Defaults($n)
    pack $f.lab $f.but -side left -padx 5
  }
}

proc remove_color {w {cnt 1}} {
  global Defaults
  set n $Defaults(nc)
  if {$n == 1} {
    errormsg "need at least one color" $w
    return
  }
  incr Defaults(nc) -$cnt
  set cf [FrameGet $w.colors]
  while {$n > $Defaults(nc)} {
    pack forget $cf.c$n
    destroy $cf.c$n
    incr n -1
  }
}

proc set_color {but n} {
  global Defaults
  set clr [select_color $Defaults($n) "Select Color" $but]
  if {$clr != ""} {
    set Defaults($n) $clr
    color_button $but $clr
  }
}

proc color_value {clr} {
  return [format "#%2.2x%2.2x%2.2x" \
    [expr int(255.0 * [lindex $clr 0])] \
    [expr int(255.0 * [lindex $clr 1])] \
    [expr int(255.0 * [lindex $clr 2])]]
}

proc color_lighten {clr} {
  set newclr ""
  foreach c $clr {
    set nc [expr $c + 0.25]
    if {$nc < 0.0} {set nc 0.0}
    if {$nc > 1.0} {set nc 1.0}
    lappend newclr $nc
  }
  return [color_value $newclr]
}

proc gray_value {clr} {
  return [expr 0.30 * [lindex $clr 0] + \
               0.59 * [lindex $clr 1] + \
               0.11 * [lindex $clr 2]]
}

proc color_button {but clr} {
  global fgColor bgColor
  if {$clr == ""} {
    $but configure -fg $fgColor(normal) -activeforeground $fgColor(active) \
      -bg $bgColor(normal) -activebackground $bgColor(active)
  } else {
    if {[gray_value $clr] > 0.5} {
      set fg black
    } else {
      set fg white
    }
    $but configure -fg $fg -activeforeground $fg \
      -bg [color_value $clr] -activebackground [color_lighten $clr]
  }
}

proc select_color {oldclr title {loc .}} {
  set clr [color_value $oldclr]
  set newclr [tk_chooseColor -initialcolor $clr -parent $loc -title $title]
  if {$newclr != ""} {
    set white [winfo rgb . "#ffffffffffff"]
    set rgb [winfo rgb . $newclr]
    set r [expr double([lindex $rgb 0]) / double([lindex $white 0])]
    set g [expr double([lindex $rgb 1]) / double([lindex $white 1])]
    set b [expr double([lindex $rgb 2]) / double([lindex $white 2])]
    set newclr [list $r $g $b]
  }
  return $newclr
}

#----- set perspective -----

set Perspective(fov) $ProgData(fov)
set Perspective(fd) [expr 0.5 / tan(0.00872665 * $Perspective(fov))]
set Perspective(np) [expr $Perspective(fd) * $ProgData(np)]
set Perspective(fp) [expr $Perspective(fd) * $ProgData(fp)]
set Perspective(ar) 1.0

set Viewpoint(x) 0
set Viewpoint(y) 0
set Viewpoint(z) $Perspective(fd)

proc apply_perspective {what val} {
  global ProgData Perspective OGLwin Viewpoint
  set Perspective(fov) $ProgData(fov)
  set fd [expr 0.5 / tan(0.00872665 * $Perspective(fov))]
  set Viewpoint(z) [expr $fd * $Viewpoint(z) / $Perspective(fd)]
  set Perspective(fd) $fd
  if {[expr $ProgData(fp) <= $ProgData(np)]} {
    if {$what == "np"} {
      set ProgData(fp) [expr 0.01 * (100.0 * $ProgData(np) + 1.0)]
    } else {
      set ProgData(np) [expr $ProgData(fp) - 0.001]
    }
  }
  set Perspective(np) [expr $fd * $ProgData(np)]
  set Perspective(fp) [expr $fd * $ProgData(fp)]
  compute_viewport $OGLwin
  set_view $OGLwin
}

proc set_perspective {{loc .}} {
  global ProgData fgColor Font

  set w .perspective
  if [winfo exists $w] {
      wm deiconify $w
      return
  }

  toplevel $w
  wm title $w "Perspective"

  FrameCreate $w.opts -font $Font(bold)
  pack $w.opts -side top -padx 5 -pady 5
  set opts [FrameGet $w.opts]

  set f [frame $opts.fov]
  pack $f -side top
  label $f.lab -width 18 -anchor w -text "Field of View"
  scale $f.scl -length 200 -from 5 -to 85 -variable ProgData(fov) \
    -orient horizontal -command {apply_perspective fov}
  pack $f.lab $f.scl -side left

  set f [frame $opts.np]
  pack $f -side top
  label $f.lab -width 18 -anchor w -text "Near Clipping Plane"
  scale $f.scl -length 200 -from 0.001 -to 1 -resolution .001 \
    -variable ProgData(np) -orient horizontal -command {apply_perspective np}
  pack $f.lab $f.scl -side left

  set f [frame $opts.fp]
  pack $f -side top
  label $f.lab -width 18 -anchor w -text "Far Clipping Plane"
  scale $f.scl -length 200 -from .01 -to 5 -resolution .01 \
    -variable ProgData(fp) -orient horizontal -command {apply_perspective fp}
  pack $f.lab $f.scl -side left

  set f [frame $w.but]
  pack $f -side top
  button $f.default -text Defaults -width 8 -command {
    array set ProgData {
      fov 30
      np 0.025
      fp 2
    }
    apply_perspective all 0
  }
  button $f.close -text Close -width 8 -command "destroy $w"
  pack $f.default $f.close -side left -pady 5

  center_window $w $loc
}

#----- OGL window controls

set ProgData(axislist) [OGLaxis $ProgData(axis)]
$OGLwin main -clear colorbuffer depthbuffer -call $ProgData(axislist)

eval $OGLwin eval \
  -clearcolor $ProgData(background) \
  -$ProgData(culling) cullface \
  -matrixmode projection \
  -loadidentity \
  -perspective $Perspective(fov) $Perspective(ar) \
    $Perspective(np) $Perspective(fp) \
  -matrixmode modelview \
  -loadidentity \
  -light light0 position 0 1 5 0 \
  -light light0 diffuse .8 .8 .8 \
  -light light0 specular 1 1 1 \
  -light light0 ambient .2 .2 .2 \
  -lightmodel lightmodeltwoside $ProgData(twosided) \
  -enable lighting -enable light0 \
  -enable depthtest -enable normalize -shademodel smooth

set ViewMatrix [$OGLwin get modelviewmatrix]

bind $OGLwin <Configure> {compute_viewport %W}
bind $OGLwin <Any-ButtonPress> {set xlast %x ; set ylast %y}
bind $OGLwin <B1-Motion> {Rotate %W %x %y}
bind $OGLwin <Shift-B1-Motion> {Pan %W %x %y}
bind $OGLwin <Control-B1-Motion> {Zoom %W %x %y}
bind $OGLwin <B2-Motion> {Zoom %W %x %y}
bind $OGLwin <B3-Motion> {Pan %W %x %y}

catch {bind . <MouseWheel> {WheelZoom %D}}
bind . <KeyPress-c> {Center}
bind . <KeyPress-r> {reset_view ; $OGLwin redraw}

proc compute_viewport {w} {
  global ProgData Perspective
  set ProgData(width) [winfo width $w]
  set ProgData(height) [winfo height $w]
  set ProgData(scale) [expr 1.0 / double($ProgData(height))]
  set Perspective(ar) [expr double($ProgData(width)) / \
      double($ProgData(height))]
  $w eval \
    -matrixmode projection \
    -loadidentity \
    -perspective $Perspective(fov) $Perspective(ar) \
      $Perspective(np) $Perspective(fp) \
    -matrixmode modelview
}

proc set_view {w} {
  global Viewpoint ViewMatrix
  eval $w eval \
    -loadidentity \
    -lookat $Viewpoint(x) $Viewpoint(y) $Viewpoint(z) \
      $Viewpoint(x) $Viewpoint(y) 0 0 1 0 \
    -multmatrix $ViewMatrix
  $w redraw
}

proc Rotate {w x y} {
  global xlast ylast ViewMatrix Viewpoint Perspective
  set scl [expr $Viewpoint(z) / $Perspective(fd)]
  if {[expr $scl > 1.0]} {
    set dx [expr $x - $xlast]
    set dy [expr $y - $ylast]
  } else {
    if {[expr $scl < 0.1]} {set scl 0.1}
    set dx [expr $scl * ($x - $xlast)]
    set dy [expr $scl * ($y - $ylast)]
  }
  set xlast $x
  set ylast $y
  eval $w eval \
    -loadidentity \
    -rotate $dy 1 0 0 \
    -rotate $dx 0 1 0 \
    -multmatrix $ViewMatrix
  set ViewMatrix [$w get modelviewmatrix]
  set_view $w
}

proc Pan {w x y} {
  global xlast ylast Viewpoint Perspective ProgData
  set scl [expr $Viewpoint(z) / $Perspective(fd)]
  if {[expr $scl > 1.0]} {
    set scl $ProgData(scale)
  } else {
    if {[expr $scl < 0.1]} {set scl 0.1}
    set scl [expr $scl * $ProgData(scale)]
  }
  set Viewpoint(x) [expr $scl * ($xlast - $x) + $Viewpoint(x)]
  set Viewpoint(y) [expr $scl * ($y - $ylast) + $Viewpoint(y)]
  set xlast $x
  set ylast $y
  set_view $w
}

proc Zoom {w x y} {
  global xlast ylast Viewpoint
  set scl [expr 0.01 * $Viewpoint(z)]
  if {[expr $scl < 0.001]} {set scl 0.001}
  set Viewpoint(z) [expr $scl * ($ylast - $y) + $Viewpoint(z)]
  if {[expr $Viewpoint(z) < 0.001]} {
    set Viewpoint(z) 0.001
  }
  set xlast $x
  set ylast $y
  set_view $w
}

proc WheelZoom {d} {
  global Viewpoint OGLwin
  set scl [expr 0.001 * $Viewpoint(z)]
  if {[expr $scl < 0.0001]} {set scl 0.0001}
  set Viewpoint(z) [expr $scl * $d + $Viewpoint(z)]
  if {[expr $Viewpoint(z) < 0.001]} {
    set Viewpoint(z) 0.001
  }
  set_view $OGLwin
}

proc reset_clipping {} {
  global ProgData Perspective OGLwin

  array set ProgData {
    np 0.025
    fp 2.0
  }
  set Perspective(np) [expr $Perspective(fd) * $ProgData(np)]
  set Perspective(fp) [expr $Perspective(fd) * $ProgData(fp)]
  compute_viewport $OGLwin
}

proc Center {} {
  global ProgData Viewpoint ViewMatrix Perspective OGLwin

  set bounds [CGNSbounds $ViewMatrix]
  set b [lindex $bounds 0]
  set xt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set xd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set b [lindex $bounds 1]
  set yt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set yd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set b [lindex $bounds 2]
  set zt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set zd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set dd [expr $xd * $xd + $yd * $yd + $zd * $zd]
  if {[expr $dd < 1.0e-15]} {
    set scl 1.0
  } else {
    set scl [expr 1.0 / sqrt($dd)]
  }

#  reset_clipping
  eval $OGLwin eval \
    -matrixmode modelview \
    -loadidentity \
    -scale $scl $scl $scl \
    -translate $xt $yt $zt \
    -multmatrix $ViewMatrix
  set ViewMatrix [$OGLwin get modelviewmatrix]

  set Viewpoint(x) 0
  set Viewpoint(y) 0
  set Viewpoint(z) $Perspective(fd)
  set_view $OGLwin
}

proc reset_view {} {
  global ProgData Viewpoint ViewMatrix Perspective OGLwin

  set bounds [CGNSbounds]
  set b [lindex $bounds 0]
  set xt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set xd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set b [lindex $bounds 1]
  set yt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set yd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set b [lindex $bounds 2]
  set zt [expr -0.5 * ([lindex $b 1] + [lindex $b 0])]
  set zd [expr double([lindex $b 1]) - double([lindex $b 0])]
  set dd [expr $xd * $xd + $yd * $yd + $zd * $zd]
  if {[expr $dd < 1.0e-15]} {
    set scl 1.0
  } else {
    set scl [expr 1.0 / sqrt($dd)]
  }

  set Viewpoint(x) 0
  set Viewpoint(y) 0
  set Viewpoint(z) $Perspective(fd)

  reset_clipping
  $OGLwin eval \
    -matrixmode modelview \
    -loadidentity \
    -scale $scl $scl $scl \
    -translate $xt $yt $zt
  set ViewMatrix [$OGLwin get modelviewmatrix]
  eval $OGLwin eval \
    -loadidentity \
    -lookat $Viewpoint(x) $Viewpoint(y) $Viewpoint(z) \
      $Viewpoint(x) $Viewpoint(y) 0 0 1 0 \
    -multmatrix $ViewMatrix
}

proc init_display {} {
  global ProgData OGLwin
  if {$ProgData(displaylist) != ""} {
    $OGLwin deletelist $ProgData(displaylist)
  }
  if {$ProgData(nzones) < 1} {
    set ProgData(displaylist) ""
    $OGLwin main -clear colorbuffer depthbuffer -call $ProgData(axislist)
  } else {
    set calllist {}
    set nz 0
    set ncolors [llength $ProgData(colors)]
    foreach z $ProgData(zones) {
      set nc [expr $nz % $ncolors]
      set clr [lindex $ProgData(colors) $nc]
      set ProgData($nz,clr) $clr
      set ProgData($nz,vis) 1
      set nr 0
      foreach r $ProgData($nz,regions) {
        if [catch {CGNSregiondim $nz $nr} dim] {
          set msg $dim
          set dim 0
        } else {
          set msg ""
        }
        if {$dim < 1} {
          set ProgData($nz,$nr,dim) 0
          set ProgData($nz,$nr,vis) 0
          set ProgData($nz,$nr,mode) 0
          set ProgData($nz,$nr,clr) ""
          if {$msg == ""} {
            set ProgData($nz,$nr,msg) "couldn't process for exterior faces"
          } else {
            set ProgData($nz,$nr,msg) $msg
          }
        } else {
          if {$nr == 0 || $dim == 3} {
            set vis $ProgData(meshvis)
            set mode $ProgData(meshmode)
          } else {
            set vis $ProgData(regvis)
            set mode $ProgData(regmode)
          }
          if {$vis} {
            set dl [OGLregion $nz $nr $mode $clr]
          } else {
            set dl [OGLregion $nz $nr 0 $clr]
          }
          lappend calllist -call $dl
          set ProgData($nz,$nr,dim) $dim
          set ProgData($nz,$nr,vis) $vis
          set ProgData($nz,$nr,mode) $mode
          set ProgData($nz,$nr,clr) $clr
          set ProgData($nz,$nr,msg) ""
        }
        incr nr
      }
      incr nz
    }
    set ProgData(displaylist) [eval $OGLwin newlist $calllist]
    $OGLwin main -clear colorbuffer depthbuffer \
      -call $ProgData(axislist) -call $ProgData(displaylist)
  }
  OGLaxis $ProgData(axis)
}

proc cleanup_display {} {
  global ProgData OGLwin

  display_message ""
  if {$ProgData(cgnsfile) == ""} return
  catch CGNSclose
  if {$ProgData(displaylist) != ""} {
    $OGLwin deletelist $ProgData(displaylist)
    set ProgData(displaylist) ""
  }
  $OGLwin main -clear colorbuffer depthbuffer -call $ProgData(axislist)
  ComboboxConfig $ProgData(baselist) -values {}
  $ProgData(zonelist) delete 0 end
  $ProgData(reglist) delete 0 end
  color_button $ProgData(zonebut) ""
  zone_state disabled
  color_button $ProgData(regionbut) ""
  region_state 0
  array set ProgData {
    nzones 0
    zones {}
    curzone ""
    curreg ""
    zoneclr ""
    zonevis 0
    regiondim 0
    regionclr ""
    regionvis 0
    regionmode 0
  }
  OGLaxis $ProgData(axis)
  $OGLwin redraw
}

#---------- set base

proc set_base {w n} {
  global ProgData

  display_message "reading zone information..."
  if {[catch {CGNSbase [expr $n + 1]} zones]} {
    display_message $zones
    errormsg $zones
    return ""
  }
  set ProgData(zones) $zones
  set ProgData(nzones) [llength $zones]
  $ProgData(zonelist) delete 0 end
  $ProgData(reglist) delete 0 end
  . configure -cursor watch
  update

  for {set n 0} {$n < $ProgData(nzones)} {incr n} {
    set nz [expr $n + 1]
    display_message "reading zone $nz..."
    if {[catch {CGNSzone $nz} regs]} {
      errormsg $regs
      set regs ""
    }
    set ProgData($n,regions) $regs
  }

  display_message "building display lists..."
  init_display
  reset_view

  if {[llength $zones] > 0} {
    foreach z $zones {
      $ProgData(zonelist) insert end $z
    }
    $ProgData(zonelist) selection set 0
    set_zone 0
  } else {
    set_zone ""
  }

  . configure -cursor {}
  display_message ""
  return ""
}

proc base_properties {w x y} {
  global Font
  if {![catch CGNSgetbase msg] && $msg != ""} {
    incr y [winfo rooty $w]
    popup_message $msg -parent $w -position "$x $y" -font $Font(fixed) -wrap 0
  }
}

#---------- set_zone

proc set_zone {nz} {
  global ProgData fgColor
  $ProgData(reglist) delete 0 end
  if {[catch {expr $nz < 0 || $nz >= $ProgData(nzones)} bad] || $bad} {
    array set ProgData {
      curzone ""
      zoneclr ""
      zonevis 0
    }
    color_button $ProgData(zonebut) ""
    zone_state disabled
    set_region ""
    return
  }
  set ProgData(curzone) $nz
  set ProgData(zonevis) $ProgData($nz,vis)
  set ProgData(zoneclr) $ProgData($nz,clr)
  zone_state normal
  color_button $ProgData(zonebut) $ProgData(zoneclr)
  set nr 0
  foreach r $ProgData($nz,regions) {
    $ProgData(reglist) insert end $r
    if {$ProgData($nz,$nr,dim)} {
      if {!$ProgData($nz,$nr,vis)} {
        catch {$ProgData(reglist) itemconfigure $nr -fg $fgColor(disabled)}
      }
    } else {
      catch {$ProgData(reglist) itemconfigure $nr -fg $ProgData(badregion)}
    }
    incr nr
  }
  if {$nr} {
    $ProgData(reglist) selection set 0
    set_region 0
  } else {
    set_region ""
  }
}

proc update_zone {} {
  global ProgData OGLwin bgColor fgColor
  set nz $ProgData(curzone)
  if {$nz == ""} {
      color_button $ProgData(zonebut) ""
      zone_state disabled
      return
  }
  set ProgData($nz,vis) $ProgData(zonevis)
  set ProgData($nz,clr) $ProgData(zoneclr)
  zone_state normal
  color_button $ProgData(zonebut) $ProgData(zoneclr)
  if {$ProgData(zonevis)} {
    set fg $fgColor(normal)
  } else {
    set fg $fgColor(disabled)
  }
  catch {$ProgData(zonelist) itemconfigure $nz -fg $fg}

  set nr 0
  foreach r $ProgData($nz,regions) {
    if {$ProgData($nz,$nr,dim)} {
      if {$ProgData(zonevis) && $ProgData($nz,$nr,vis)} {
        set mode $ProgData($nz,$nr,mode)
        set fg $fgColor(normal)
      } else {
        set mode 0
        set fg $fgColor(disabled)
      }
      catch {$ProgData(reglist) itemconfigure $nr -fg $fg}
      OGLregion $nz $nr $mode $ProgData($nz,$nr,clr)
    }
    incr nr
  }
  OGLaxis $ProgData(axis)
  $OGLwin redraw
}

proc toggle_zone {n} {
  global ProgData
  set ProgData(zonevis) [expr $ProgData(zonevis) ^ 1]
  update_zone
}

proc zone_color {w} {
  global ProgData
  set nz $ProgData(curzone)
  if {$nz == "" || $ProgData(zoneclr) == ""} return
  set clr [select_color $ProgData(zoneclr) "Zone Color" $w]
  if {$clr != ""} {
    set ProgData(zoneclr) $clr
    set nr 0
    foreach r $ProgData($nz,regions) {
      if {$ProgData($nz,$nr,dim)} {
        set ProgData($nz,$nr,clr) $clr
      }
      incr nr
    }
    if {$ProgData(curreg) != "" && $ProgData(regiondim)} {
      set ProgData(regionclr) $clr
      color_button $ProgData(regionbut) $clr
    }
    update_zone
  }
}

proc zone_properties {w x y} {
  global ProgData Font
  if {!$ProgData(nzones)} return
  set nz [$w nearest $y]
  if {$nz == ""} return
  catch {CGNSgetzone $nz} msg
  if {$msg != ""} {
    incr y [winfo rooty $w]
    popup_message $msg -parent $w -position "$x $y" -font $Font(fixed) -wrap 0
  }
}

#---------- set region

proc set_region {regs} {
  global ProgData
  set nz $ProgData(curzone)
  set ProgData(curreg) $regs
  set nreg [llength $regs]
  if {$nreg == 0} {
    array set ProgData {
      regiondim 0
      regionvis 0
      regionmode 0
      regionclr ""
    }
    display_message ""
  } else {
    set nr [lindex $regs 0]
    set ProgData(regiondim) $ProgData($nz,$nr,dim)
    set ProgData(regionvis) $ProgData($nz,$nr,vis)
    set ProgData(regionmode) $ProgData($nz,$nr,mode)
    set ProgData(regionclr) $ProgData($nz,$nr,clr)
    if {$nreg == 1} {
      display_message $ProgData($nz,$nr,msg)
    } else {
      display_message ""
      for {set n 1} {$n < $nreg} {incr n} {
        set nr [lindex $regs $n]
        if {$ProgData($nz,$nr,dim)} {
          set Progdata(regiondim) 1
        }
        if {!$ProgData($nz,$nr,vis)} {
          set ProgData(regionvis) 0
        }
        if {$ProgData($nz,$nr,mode) != $ProgData(regionmode)} {
          set ProgData(regionmode) 0
        }
        if {$ProgData($nz,$nr,clr) != $ProgData(regionclr)} {
          set ProgData(regionclr) ""
        }
      }
    }
  }
  color_button $ProgData(regionbut) $ProgData(regionclr)
  region_state $ProgData(regiondim)
}

proc toggle_region {n} {
  global ProgData
  if {$ProgData(regiondim)} {
    set ProgData(regionvis) [expr $ProgData(regionvis) ^ 1]
    update_region
  }
}

proc update_region {} {
  global ProgData OGLwin bgColor fgColor
  set nz $ProgData(curzone)
  set regs $ProgData(curreg)
  if {$nz == "" || $regs == ""} {
    color_button $ProgData(regionbut) ""
    region_state 0
    return
  }
  foreach nr $regs {
    if {$ProgData($nz,$nr,dim)} {
      set ProgData($nz,$nr,vis) $ProgData(regionvis)
      if {$ProgData(regionmode)} {
        set ProgData($nz,$nr,mode) $ProgData(regionmode)
      }
      if {$ProgData(regionclr) != ""} {
        set ProgData($nz,$nr,clr) $ProgData(regionclr)
      }
      if {$ProgData(zonevis) && $ProgData(regionvis)} {
        set mode $ProgData($nz,$nr,mode)
        set fg $fgColor(normal)
      } else {
        set mode 0
        set fg $fgColor(disabled)
      }
      catch {$ProgData(reglist) itemconfigure $nr -fg $fg}
      OGLregion $nz $nr $mode $ProgData($nz,$nr,clr)
    }
  }
  region_state $ProgData(regiondim)
  color_button $ProgData(regionbut) $ProgData(regionclr)
  OGLaxis $ProgData(axis)
  $OGLwin redraw
}

proc region_color {w} {
  global ProgData
  if {!$ProgData(regiondim)} return
  if {$ProgData(regionclr) == ""} {
    set oldclr $ProgData(zoneclr)
  } else {
    set oldclr $ProgData(regionclr)
  }
  set newclr [select_color $oldclr "Region Color" $w]
  if {$newclr != ""} {
    set ProgData(regionclr) $newclr
    update_region
  }
}

proc region_properties {w x y} {
  global ProgData Font
  set nz $ProgData(curzone)
  set nr [$w nearest $y]
  if {$nz == "" || $nr == ""} return
  catch {CGNSgetregion $nz $nr} msg
  if {$msg != ""} {
    incr y [winfo rooty $w]
    popup_message $msg -parent $w -position "$x $y" -font $Font(fixed) -wrap 0
  }
}

#---------- set variable for plotting

proc set_variable {w n} {
  global ProgData
  return ""
}

#---------- read cgns file

set FileList {
  {{CGNS Files} {.cgns .cgn .adf}}
  {{All Files} {*}}
}

proc load_cgns {{filename ""}} {
  global ProgData FileList ModelData
  global Translate Scale

  if {$filename == ""} {
    set filename [FileOpen "Open CGNS File" \
      $ProgData(cgnsfile) . $FileList]
  }
  if {$filename == ""} return
  cleanup_display
  display_message "reading base information..."
  . configure -cursor watch
  update
  set msg ""
  if {[catch {CGNSopen $filename} bases]} {
    set msg $bases
  } else {
    set msg ""
  }
  . configure -cursor {}
  if {$msg != ""} {
    display_message $msg
    errormsg $msg
    return
  }
  ComboboxConfig $ProgData(baselist) -values $bases -index 0
  set_base $ProgData(baselist) 0
  if {[llength $bases] > 1} {
    ComboboxConfig $ProgData(baselist) -state normal
  } else {
    ComboboxConfig $ProgData(baselist) -state disabled
  }

  wm title . "CGNSplot : [file tail $filename]"
  set ProgData(cgnsfile) $filename
}

if {[info commands etLoadIcon] != ""} {
  wm withdraw .
  update idletasks
  if [catch {etLoadIcon cgnsplot}] {
    catch {etLoadIcon cgns}
  }
  wm deiconify .
}

if {$argc} {
  set file [lindex $argv [expr $argc - 1]]
  if {[string index $file 0] != "-" && [file exists $file]} {
    load_cgns $file
  }
}

